@@ -17,6 +17,15 @@ from .solver.solver import solver
 from ..branch_addition.basis import tangent_basis
 from pickle import dumps,loads
 from scipy.spatial import cKDTree
+from scipy.sparse import csr_matrix
+from scipy.sparse.csgraph import shortest_path
+from scipy.interpolate import splprep,splev
+from gekko import GEKKO
+import marshal, types
+
+import pyvista as pv
+import tetgen
+import time
 
 import imp
 import sys
@@ -67,11 +76,14 @@ class patch:
 
 class surface:
     def __init__(self):
-        self.other_surface = None
+        self.subtracted_volumes = []
+        self.added_volumes = []
+        self.polydata = None
         pass
-    def load(self,filename):
-        points,normals = load3d(filename)
+    def load(self,filename,subdivisions=0):
+        points,normals,polydata = load3d(filename,subdivisions)
         self.set_data(points,normals=normals)
+        self.polydata = polydata
     def set_data(self,points,normals=None,workers=1,local_min=10,local_max=20,l=0.5,PCA_samplesize=25):
         self.points = points
         self.ddim = points.shape[1]
@@ -252,7 +264,15 @@ class surface:
             self.max_patch_size = self.points.shape[0]
             self.patches = self.parallel_build(None,show_individual=show_individual,PU=PU)
 
-    def build(self,h=0,q=1,d_num=2):
+    def build(self,**kwargs):
+        h          = kwargs.get('h',0)
+        q          = kwargs.get('q',1)
+        d_num      = kwargs.get('d_num',2)
+        resolution = kwargs.get('resolution',20)
+        k          = kwargs.get('k',2)
+        level      = kwargs.get('level',0)
+        visualize  = kwargs.get('visualize',False)
+        buffer     = kwargs.get('buffer',1)
         self.x_range = [min(self.points[:,0]),max(self.points[:,0])]
         self.y_range = [min(self.points[:,1]),max(self.points[:,1])]
         self.z_range = [min(self.points[:,2]),max(self.points[:,2])]
@@ -316,22 +336,7 @@ class surface:
                                     a_coef=a_coef,b_coef=b_coef,h=h,q=q,
                                     patch_x=np.array(patch_x),patch_y=np.array(patch_y),
                                     patch_z=np.array(patch_z))
-            #ls_d0 = partial(linesearch_d1,patch_points=np.array(patch_points),
-            #                b_coef=np.array(b_sol),h=h,q=q,sol_mat=np.array(solution_matrix),
-            #                patch_1=np.array(patch_x),patch_2=np.array(patch_y),
-            #                patch_3=np.array(patch_z))
-            #ls_d1 = partial(linesearch_d1,patch_points=np.array(patch_points),
-            #                b_coef=np.array(b_sol),h=h,q=q,sol_mat=np.array(solution_matrix),
-            #                patch_1=np.array(patch_x),patch_2=np.array(patch_y),
-            #                patch_3=np.array(patch_z))
-            #ls_d2 = partial(linesearch_d2,patch_points=np.array(patch_points),
-            #                b_coef=np.array(b_sol),h=h,q=q,sol_mat=np.array(solution_matrix),
-            #                patch_1=np.array(patch_x),patch_2=np.array(patch_y),
-            #                patch_3=np.array(patch_z))
             self.function_marching = func_marching
-            #self.linesearch_d0 = ls_d0
-            #self.linesearch_d1 = ls_d1
-            #self.linesearch_d2 = ls_d2
         else:
             functions = []
             preassembled_functions,function_strings = construct_global(d_num)
@@ -350,7 +355,8 @@ class surface:
                 exec("DD.append(partial(foo.{},patch_points=np.array(patch_points),b_coef=np.array(b_sol),sol_mat=np.array(solution_matrix)))".format(fn))
         self.pickled_DD = pickled_DD
         self.DD = DD
-        self._get_properties_()
+        #print('get properties')
+        self._get_properties_(resolution=resolution,k=k,level=level,visualize=visualize,buffer=buffer)
         """
         DD = []
         CDD = []
@@ -385,34 +391,58 @@ class surface:
         self.CDD = CDD
         """
     def within(self,x,y,z,k):
-        if self.other_surface is None:
-            return self.DD[0]([x,y,z,k]) < 0
-        else:
-            return (self.DD[0]([x,y,z,k]) < 0 and self.other_surface.DD[0]([x,y,z,k]) > 0)
+        return self.DD[0]([x,y,z,k]) < 0
+
+    def within_mesh(self,x,y,z,k):
+        point = pv.PolyData(np.array([x,y,z]))
+        select = point.select_enclosed_points(self.pv_polydata)
+        return bool(select['SelectedPoints'][0])
+
+    def pick_in_cell(self,cell_idx):
+        simplex = self.tet_pts[self.tet_verts[cell_idx,:],:]
+        point = np.array(generate(simplex,1))
+        return point,None
 
     def pick(self,**kwargs):
         homogeneous = kwargs.get('homogeneous',False)
+        replacement = kwargs.get('replacement',True)
+        size = kwargs.get('size',1)
+        vert_applied = kwargs.get('vert',False)
+        verts = kwargs.get('verts',[])
         if homogeneous:
             point = None
-            while point is None:
-                x = np.random.uniform(self.x_range[0],self.x_range[1])
-                y = np.random.uniform(self.y_range[0],self.y_range[1])
-                z = np.random.uniform(self.z_range[0],self.z_range[1])
-                if self.within(x,y,z,2):
-                    point = np.array([x,y,z])
-                    break
+            if not vert_applied:
+                rdx = np.random.Generator(np.random.PCG64()).choice(list(range(self.tet_verts.shape[0])),size,p=self.norm_ele_vol,replace=replacement)
+            else:
+                rdx = verts
+            simplex = self.tet_pts[self.tet_verts[rdx,:],:]
+            point = np.array(generate(simplex,size))
             return point,None
-        x = np.random.uniform(self.x_range[0],self.x_range[1])
-        y = np.random.uniform(self.y_range[0],self.y_range[1])
-        z = np.random.uniform(self.z_range[0],self.z_range[1])
-        k     = kwargs.get('k',len(self.patches))
-        layer = kwargs.get('layer',0)
-        x0    = kwargs.get('x0',np.array([x,y,z]))
-        f     = lambda x: (self.DD[0]([x[0],x[1],x[2],k])-layer)**2
-        df    = lambda x: 2*(self.DD[0]([x[0],x[1],x[2],k])-layer)*self.DD[1]([x[0],x[1],x[2],k])
+        id = np.random.choice(self.pv_polydata_surf.n_faces)
+        faces = self.pv_polydata_surf.faces.reshape(-1,self.pv_polydata_surf.faces[0]+1)
+        pts = self.pv_polydata_surf.points[faces[id,1:],:]
+        a1 = np.random.random(1)
+        a2 = np.random.random(1)
+        if (a1+a2) > 1:
+            a1 = 1 - a1
+            a2 = 1 - a2
+        A = 1 - a1 - a2
+        B = a1
+        C = a2
+        pts = pts*np.array([A,B,C])
+        pts = np.sum(pts,axis=0)
+        pts = np.array(pts)
+        #x = np.random.uniform(self.x_range[0],self.x_range[1])
+        #y = np.random.uniform(self.y_range[0],self.y_range[1])
+        #z = np.random.uniform(self.z_range[0],self.z_range[1])
+        #k     = kwargs.get('k',len(self.patches))
+        #layer = kwargs.get('layer',0)
+        #x0    = kwargs.get('x0',np.array([x,y,z]))
+        #f     = lambda x: (self.DD[0]([x[0],x[1],x[2],k])-layer)**2
+        #df    = lambda x: 2*(self.DD[0]([x[0],x[1],x[2],k])-layer)*self.DD[1]([x[0],x[1],x[2],k])
         #ddf   = lambda x,y,z: 2*np.dot(self.DD[1]([x,y,z,k],self.DD[1]([x,y,z,k]))+2*self.DD[2]([x,y,z,k])*(self.DD[0]([x,y,z,k])-layer)
-        res   = optimize.minimize(f,x0,method='L-BFGS-B')
-        return res.x,res.fun
+        #res   = optimize.minimize(f,x0,method='L-BFGS-B')
+        return pts,None
 
     def path(self,start,distance,steps,dive=0.01,theta_steps=40):
         start_layer = self.DD[0]([start[0],start[1],start[2],len(self.patches)])
@@ -535,43 +565,349 @@ class surface:
         k          = kwargs.get('k',2)
         level      = kwargs.get('level',0)
         visualize  = kwargs.get('visualize',False)
-        self.polydata   = marching_cubes(self,resolution)
+        buffer     = kwargs.get('buffer',1)
+        self.polydata,self.tet_pts,self.tet_verts,self.ele_vol,self.tet = marching_cubes(self,resolution,k,level,visualize,buffer)
+        self.cell_lookup = cKDTree(self.tet.grid.cell_centers().points)
+        self.pv_polydata = pv.PolyData(var_inp=self.polydata)
+        self.pv_polydata_surf = self.pv_polydata.extract_surface()
+        nodes_check = set([])
+        nodes = []
+        idx = [[0,1],[1,2],[2,3],[3,1]]
+        lengths = []
+        for tet in self.tet_verts:
+            for dx in idx:
+                if tuple([tet[dx[0]],tet[dx[1]]]) in nodes_check:
+                    continue
+                nodes_check.add(tuple([tet[dx[0]],tet[dx[1]]]))
+                nodes_check.add(tuple([tet[dx[1]],tet[dx[0]]]))
+                nodes.append([tet[dx[0]],tet[dx[1]]])
+                nodes.append([tet[dx[1]],tet[dx[0]]])
+                l = np.linalg.norm(self.tet_pts[tet[dx[0]]] - self.tet_pts[tet[dx[1]]])
+                lengths.append(l)
+                lengths.append(l)
+        M = np.array(nodes)
+        graph = csr_matrix((lengths,(M[:,0],M[:,1])),shape=(max(M[:,0])+1,max(M[:,1])+1))
+        self.graph = graph
+        self.shortest_algorithm = lambda ind: shortest_path(csgraph=graph,directed=False,indices=ind,return_predecessors=True)
+        def get_path(i,j):
+            ds, Pr = shortest_path(csgraph=graph,directed=False,indices=i,return_predecessors=True)
+            path = [j]
+            lengths = []
+            k = j
+            while Pr[k] != -9999:
+                path.append(Pr[k])
+                lengths.append(ds[k])
+                k = Pr[k]
+            #lengths.append(ds[-9999])
+            path = path[::-1]
+            lengths = lengths[::-1]
+            lines = []
+            for jdx in range(len(path)-1):
+                lines.append([path[jdx],path[jdx+1]])
+            return path,lengths,lines
+        def get_cells(path,grid=self.tet.grid):
+            cells = []
+            cell_sets = []
+            for idx,i in enumerate(path):
+                cells.append(grid.extract_points(i))
+                cell_sets.append(set(cells[-1].cell_data['vtkOriginalCellIds']))
+                if idx > 0:
+                    cell_sets[-2] = cell_sets[-2].difference(cell_sets[-1])
+                    cell_sets[-1] = cell_sets[-1].difference(cell_sets[-2])
+            cell_sets = [list(cs) for cs in cell_sets]
+            return cell_sets
+        def get_cell_points(cells,verts=self.tet_verts,pts=self.tet_pts):
+            return pts[verts[cells]]
+        def point_function(s1,s2,s3,n,cell_pts):
+            dn = 1/cell_pts.shape[0]
+            N  = int(n // dn)
+            if N > cell_pts.shape[0]-1:
+                N =  cell_pts.shape[0]-1
+            if N < 0:
+                N = 0
+            if s1 + s2 > 1:
+                s1 = 1.0 - s1
+                s2 = 1.0 - s2
+            if s2 + s3 > 1:
+                tmp = s3
+                s3 = 1.0 - s1 - s2
+                s2 = 1.0 - tmp
+            elif s1 + s2 + s3 > 1:
+                tmp = s3
+                s3 = s1 + s2 + s3 - 1.0
+                s1 = 1.0 - s2 - tmp
+            s4 = 1 - s1 - s2 - s3
+            #print(s4)
+            #print(s1.VALUE)
+            #print(s2.VALUE)
+            #print(s3.VALUE)
+            #print(s4.value)
+            S  = np.array([s1,s2,s3,s4]).reshape(-1,1)
+            #id = 0
+            #for i in range(cell_pts.shape[0]):
+            #    if n.value == i:
+            #        id = i
+            #        break
+            #P1  = s1*cell_pts[id,0,0] + s2*cell_pts[id,1,0] + s3*cell_pts[id,2,0] + s4*cell_pts[id,3,0]
+            #P2  = s1*cell_pts[id,0,1] + s2*cell_pts[id,1,1] + s3*cell_pts[id,2,1] + s4*cell_pts[id,3,1]
+            #P3  = s1*cell_pts[id,0,2] + s2*cell_pts[id,1,2] + s3*cell_pts[id,2,2] + s4*cell_pts[id,3,2]
+            #print(id)
+            P  = cell_pts[N]*S
+            return np.sum(P,axis=0)
+
+        def get_point_functions(cell_sets):
+            point_functions = []
+            for i in range(len(cell_sets)):
+                cell_pts = get_cell_points(cell_sets[i])
+                f = partial(point_function,cell_pts=cell_pts)
+                point_functions.append(f)
+            return point_functions
+        def path_function(x,start,end,point_functions):
+            a = [start]
+            b = []
+            #l = []
+            for i in range(len(point_functions)):
+                x_tmp = x[i*4:i*4+4]
+                pt = point_functions[i](x_tmp[0],x_tmp[1],x_tmp[2],x_tmp[3])
+                #l.append(((pt[0] - a[-1][0])**2+(pt[1] - a[-1][1])**2+(pt[2] - a[-1][2])**2)**(1/2))
+                a.append(pt)
+                b.append(pt)
+            b.append(end)
+            #l.append(((pt[0] - end[0])**2+(pt[1] - end[1])**2+(pt[2] - end[2])**2)**(1/2))
+            A = np.array(a)
+            B = np.array(b)
+            L = np.sum(np.linalg.norm(B-A,axis=1))
+            #L = sum(l)
+            return L
+        def build_path_function(i,j,start,end):
+            path,_,_ = get_path(i,j)
+            cell_sets = get_cells(path)
+            point_functions = get_point_functions(cell_sets)
+            pf = partial(path_function,start=start,end=end,point_functions=point_functions)
+            return pf,point_functions
+        def optimize_path(pf,func_num,niter=200,disp=True):
+            x0 = np.zeros(func_num*4)
+            bounds = []
+            for i in range(len(x0)):
+                bounds.append(tuple([0,0.99]))
+            kwargs = {'method':"L-BFGS-B",
+                      'bounds':bounds}
+            #m = GEKKO(remote=False)
+            #x = []
+            #for i in range(len(funcs)):
+            #    x.append(m.Var(value=0,lb=0,ub=1))
+            #    x.append(m.Var(value=0,lb=0,ub=1))
+            #    x.append(m.Var(value=0,lb=0,ub=1))
+            #    x.append(m.Var(value=0,lb=0,ub=len(funcs[i].keywords['cell_pts']),integer=True))
+            #x = m.Array(m.Var,len(x0),lb=0,ub=1,integer=True)
+            #m.Minimize(pf(x))
+            #m.options.SOLVER=1
+            #m.solver_options = ['minlp_maximum_iterations 500', \
+            #        # minlp iterations with integer solution
+            #        'minlp_max_iter_with_int_sol 10', \
+            #        # treat minlp as nlp
+            #        'minlp_as_nlp 0', \
+            #        # nlp sub-problem max iterations
+            #        'nlp_maximum_iterations 50', \
+            #        # 1 = depth first, 2 = breadth first
+            #        'minlp_branch_method 1', \
+            #        # maximum deviation from whole number
+            #        'minlp_integer_tol 0.05', \
+            #        # covergence tolerance
+            #        'minlp_gap_tol 0.01']
+            #m.solve()
+
+            res = optimize.basinhopping(pf,x0,niter=niter,minimizer_kwargs=kwargs,disp=disp)
+            return res
+        def optimized_points(res,point_functions,start,end):
+            pts = [start]
+            a = [start]
+            b = []
+            sol = res.x
+            for i in range(len(point_functions)):
+                x_tmp = sol[i*4:i*4+4]
+                pts.append(point_functions[i](x_tmp[0],x_tmp[1],x_tmp[2],x_tmp[3]))
+                a.append(pts[-1])
+                b.append(pts[-1])
+            pts.append(end)
+            b.append(end)
+            A = np.array(a)
+            B = np.array(b)
+            L = np.linalg.norm(B-A,axis=1)
+            return pts,L
+        def find_best_path(start,end,grid=self.tet.grid,niter=200,disp=True):
+            i = grid.find_closest_point(start)
+            j = grid.find_closest_point(end)
+            pf,funcs = build_path_function(i,j,start,end)
+            res = optimize_path(pf,len(funcs),niter=niter,disp=disp)
+            pts,L = optimized_points(res,funcs,start,end)
+            ## check for unique
+            _,u_pts_idx = np.unique(pts,axis=0,return_index=True)
+            u_pts_idx = np.flip(u_pts_idx)
+            rm_idx = np.argwhere(np.diff(u_pts_idx)>1)
+            return pts,L,res,pf,funcs,rm_idx
+        def spline_resample(path_pts):
+            if len(path_pts) == 2:
+                path = splprep(path_pts.T,k=1,s=0)
+            elif len(path_pts) == 3:
+                path = splprep(path_pts.T,k=2,s=0)
+            else:
+                path = splprep(path_pts.T,s=0)
+            return path
+        def resample(path,niter=20,disp=True):
+            def resample_optimizer(on_off,path_pts=path,within=self.pickled_DD[0]):
+                on_off = on_off > 0.5
+                last_pts = path_pts[0]
+                end_pts = path_pts[-1]
+                cost = 0
+                within = loads(within)
+                cost += np.sum(on_off)
+                for idx,i in enumerate(on_off):
+                    if i == 1:
+                        subdivision = 4
+                        for j in range(1,2*subdivision):
+                            tmp = (last_pts + path_pts[idx+1])*(j/(2*subdivision))
+                            if within((tmp[0],tmp[1],tmp[2],2)) > 0.15:
+                                cost += 100
+                        last_pts = path_pts[idx+1]
+                subdivision  = 4
+                for j in range(1,2*subdivision):
+                    tmp = (last_pts + path_pts[-1])*(j/(2*subdivision))
+                    if within((tmp[0],tmp[1],tmp[2],2)) > 0.15:
+                        cost += 100
+                return cost
+            x0 = np.ones(len(path)-2)
+            bounds = []
+            for i in range(len(x0)):
+                bounds.append(tuple([0,1]))
+            kwargs = {'method':"L-BFGS-B",
+                      'bounds':bounds}
+            #print(len(path))
+            #m = GEKKO(remote=False)
+            #x = m.Array(m.Var,len(x0),lb=0,ub=1,integer=True)
+            #m.Minimize(resample_optimizer(x))
+            #m.options.SOLVER=1
+            #m.solve()
+            #on_off = x
+            #print(on_off)
+            res = optimize.basinhopping(resample_optimizer,x0,niter=100,minimizer_kwargs=kwargs,disp=disp)
+            on_off = res.x > 0.5
+            #print(res[1])
+            new_path = [path[0]]
+            new_lengths = []
+            for idx,i in enumerate(on_off):
+                if i:
+                    new_path.append(path[idx+1])
+                    new_lengths.append(np.linalg.norm(new_path[-1]-new_path[-2]))
+            new_path.append(path[-1])
+            new_lengths.append(np.linalg.norm(new_path[-1]-new_path[-2]))
+            #print(len(new_path))
+            return np.array(new_path), new_lengths
+        self.get_shortest_path = get_path
+        self.build_path_function = build_path_function
+        self.optimize_path = optimize_path
+        self.optimized_points = optimized_points
+        self.find_best_path = find_best_path
+        self.resample = resample
+        self.total_ele_vol = sum(self.ele_vol)
+        self.norm_ele_vol = self.ele_vol/self.total_ele_vol
         volume,surface_area = properties(self.polydata)
         self.volume= volume
         self.surface_area = surface_area
 
     def subtract(self,other_surface):
-        self.other_surface = other_surface
+        self.subtracted_volumes.append(other_surface)
+        def d_0(xyz,F=self.DD[0],S=other_surface.DD[0]):
+            return F(xyz)-(S(xyz)<0)*(F(xyz)+S(xyz))
+        self.DD[0] = d_0
+        def d_1(xyz,F=self.DD[0],S=other_surface.DD[0],dF=self.DD[1],dS=other_surface.DD[1]):
+            return dF(xyz)-(S(xyz)<0)*(1-S(xyz))*(dF(xyz)+dS(xyz))
+        self.DD[1] = d_1
+        def d_2(xyz,F=self.DD[0],S=other_surface.DD[0],ddF=self.DD[2],ddS=self.DD[2]):
+            return ddF(xyz)-(S(xyz)<0)*(1-S(xyz))*(ddF(xyz)+ddS(xyz))
+        self.DD[2] = d_2
+        self.volume = self.volume - other_surface.volume
+        self.surface_area = self.surface_area - other_surface.surface_area
 
-def marching_cubes(surface_object,resolution=20,k=2,level=0,visualize=False):
+    def add(self,other_surface):
+        self.added_volumes.append(other_surface)
+        def d_0(xyz,F=self.DD[0],S=other_surface.DD[0]):
+            return F(xyz)+(S(xyz)<0)(F(xyz)>0)*(S(xyz))
+        self.DD[0] = d_0
+        def d_1(xyz,F=self.DD[0],S=other_surface.DD[0],dF=self.DD[1],dS=other_surface.DD[1]):
+            return dF(xyz)+(S(xyz)<0)*(F(xyz)>0)*dS(xyz)
+        self.DD[1] = d_1
+        def d_2(xyz,F=self.DD[0],S=other_surface.DD[0],ddF=self.DD[2],ddS=self.DD[2]):
+            return ddF(xyz)+(S(xyz)<0)*(F(xyz)>0)*ddS(xyz)
+        self.DD[2] = d_2
+
+    def export_polydata(self,file):
+        writer = vtk.vtkXMLPolyDataWriter()
+        writer.SetFileName(file)
+        writer.SetInputDataObject(self.polydata)
+        writer.Write()
+
+def marching_cubes(surface_object,resolution=20,k=2,level=0,visualize=False,buffer=1):
     """
     Takes an interpolated volume and performs
     a descritization on a selected hyperplane.
     """
     surface_algorithm = vtk.vtkMarchingCubes()
     image_data        = vtk.vtkImageData()
-    X,Y,Z = np.mgrid[surface_object.x_range[0]-1:surface_object.x_range[1]+1:resolution*1j,
-                     surface_object.y_range[0]-1:surface_object.y_range[1]+1:resolution*1j,
-                     surface_object.z_range[0]-1:surface_object.z_range[1]+1:resolution*1j]
+    colors = vtk.vtkNamedColors()
+    X,Y,Z = np.mgrid[surface_object.x_range[0]-buffer:surface_object.x_range[1]+buffer:resolution*1j,
+                     surface_object.y_range[0]-buffer:surface_object.y_range[1]+buffer:resolution*1j,
+                     surface_object.z_range[0]-buffer:surface_object.z_range[1]+buffer:resolution*1j]
+    #print(X.shape[1],X.shape[0],X.shape[2])
     SHAPE = (resolution,resolution,resolution)
-    SPACING = ((surface_object.x_range[1]-surface_object.x_range[0]+2)/(resolution-1),
-               (surface_object.y_range[1]-surface_object.y_range[0]+2)/(resolution-1),
-               (surface_object.z_range[1]-surface_object.z_range[0]+2)/(resolution-1))
+    Xs = ((-(surface_object.x_range[0]-buffer)+(surface_object.x_range[1]+buffer))/(resolution-1))
+    Ys = ((-(surface_object.y_range[0]-buffer)+(surface_object.y_range[1]+buffer))/(resolution-1))
+    Zs = ((-(surface_object.z_range[0]-buffer)+(surface_object.z_range[1]+buffer))/(resolution-1))
+    SPACING = (Zs,Ys,Xs)
     Xf = X.flatten()
     Yf = Y.flatten()
     Zf = Z.flatten()
     Kf = np.ones(Xf.shape,dtype=int)*k
+    ORIGIN = (min(Zf),min(Yf),min(Xf))
     #Sample Implicit Function
-    print('Sampling Implicit Volume...')
+    #print('Sampling Implicit Volume...')
     Vf = []
-    for i in zip(Xf,Yf,Zf,Kf):
+    for i in zip(Zf,Yf,Xf,Kf):
         Vf.append(surface_object.DD[0](i))
-    print('Converting to VTK Image Data...')
-    data_vtk = numpy_support.numpy_to_vtk(Vf,deep=1,array_type=vtk.VTK_FLOAT)
+    Vf = np.array(Vf)
+    Vf = Vf.reshape(X.shape)
+    #grid = pv.UniformGrid(dims=SHAPE,spacing=SPACING,origin=ORIGIN)
+    #out = grid.contour(1,scalars=Vf,rng=[0,0],method='marching_cubes')
+    #out.plot(color='tan',smooth_shading=True)
+    #print('Converting to VTK Image Data...')
+    #Vf = np.array(Vf)
+    #Vf = Vf.reshape(X.shape)
+    pointsdata = vtk.vtkPoints()
+    for i in range(surface_object.points.shape[0]):
+        id = pointsdata.InsertNextPoint(surface_object.points[i,0],surface_object.points[i,1],surface_object.points[i,2])
+    pt_poly = vtk.vtkPolyData()
+    pt_poly.SetPoints(pointsdata)
+    vertexfilter = vtk.vtkVertexGlyphFilter()
+    vertexfilter.SetInputData(pt_poly)
+    vertexfilter.Update()
+    PTS = vtk.vtkPolyData()
+    PTS.ShallowCopy(vertexfilter.GetOutput())
+    pt_mapper = vtk.vtkPolyDataMapper()
+    pt_mapper.SetInputData(PTS)
+    pt_actor = vtk.vtkActor()
+    pt_actor.SetMapper(pt_mapper)
+    pt_actor.GetProperty().SetPointSize(10)
+    pt_actor.GetProperty().SetColor(colors.GetColor3d('Red'))
+    #Vf = np.flipud(Vf)
+    #Vf = np.flip(Vf,1)
+    data_vtk = numpy_support.numpy_to_vtk(Vf.ravel(),deep=1,array_type=vtk.VTK_FLOAT)
     image_data.SetDimensions(SHAPE)
     image_data.SetSpacing(SPACING)
+    image_data.SetOrigin(ORIGIN)
+    #image_data.SetOrigin(-np.max(Xf)/2,np.max(Yf)/2,-np.max(Zf)/2)
     image_data.GetPointData().SetScalars(data_vtk)
-    print('Marching...')
+    image_data.Modified()
+    #print('Marching...')
     surface_algorithm.SetInputData(image_data)
     surface_algorithm.ComputeNormalsOn()
     surface_algorithm.SetValue(0,level)
@@ -599,15 +935,75 @@ def marching_cubes(surface_object,resolution=20,k=2,level=0,visualize=False):
         actor.GetProperty().SetColor(colors.GetColor3d('MistyRose'))
 
         renderer.AddActor(actor)
-
+        renderer.AddActor(pt_actor)
         render_window.Render()
         interactor.Start()
+    else:
+        pass
+        #surface_algorithm.Update()
+    if surface_object.polydata is not None:
+        volume_polydata = surface_object.polydata
     else:
         surface_algorithm.Update()
-    return surface_algorithm.GetOutput()
+        volume_polydata = surface_algorithm.GetOutput()
+    transform = vtk.vtkTransform()
+    pv_tmp = pv.PolyData(var_inp=volume_polydata)
+    if not pv_tmp.is_all_triangles:
+        #print('Triangulating mesh...')
+        pv_tmp.triangulate(inplace=True)
+        pv_tmp.plot(show_edges=True,opacity=0.5)
+    tet = tetgen.TetGen(pv_tmp)
+    try:
+        nodes, verts = tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
+    except:
+        tet.make_manifold()
+        nodes, verts = tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
+    cells = tet.grid.compute_cell_sizes()
+    ele_vol = cells.cell_data["Volume"]
+    return volume_polydata, nodes, verts, ele_vol, tet
 
 def properties(polydata_object):
     MASS = vtk.vtkMassProperties()
     MASS.SetInputData(polydata_object)
     MASS.Update()
     return MASS.GetVolume(),MASS.GetSurfaceArea()
+
+@nb.jit(nopython=True)
+def generate(points,num):
+    #r1,r2,r3 = np.random.random((3,num))
+    s = np.zeros(num*4)
+    for i in range(num):
+        r1,r2,r3 = np.random.random(3)
+        if r1 + r2 > 1:
+            r1 = 1.0 - r1
+            r2 = 1.0 - r2
+        if r2 + r3 > 1:
+            tmp = r3
+            r3 = 1.0 - r1 - r2
+            r2 = 1.0 - tmp
+        elif r1 + r2 + r3 > 1:
+            tmp = r3
+            r3 = r1 + r2 + r3 - 1.0
+            r1 = 1.0 - r2 - tmp
+        a = 1.0 - r1 - r2 - r3
+        idx_1 = i*4
+        idx_2 = i*4+1
+        idx_3 = i*4+2
+        idx_4 = i*4+3
+        s[idx_1] = a
+        s[idx_2] = r1
+        s[idx_3] = r2
+        s[idx_4] = r3
+    s = s.reshape((-1,1))
+    points = s*points.reshape(num*4,3)
+    points = np.sum(points.reshape(num,4,3),axis=1)
+    return points
+"""
+@nb.jit(nopython=True)
+def lines_to_cells(p0,centers,subdivisions):
+    points = []
+    for i in p0:
+        for j in centers:
+            for k in range(1,2*subdivisions):
+                points.append((i+j)*(k/(2*subdivisions)))
+"""
