@@ -11,8 +11,9 @@ from skimage import measure
 from time import time
 from tqdm import tqdm
 from pickle import loads
-import pandas as pd
+#import pandas as pd
 import plotly.graph_objects as go
+import vtk
 
 from ..tests.bumpy_sphere import bumpy_sphere
 
@@ -403,6 +404,30 @@ def newton(mpu_object):
     print('F1: {}'.format(f1))
     print('F2: {}'.format(f2))
     print('Ap: {}'.format(f1+np.dot(np.array([0.1,0,0]),g)))
+
+def show_patches(mpu_object):
+    patch_sets = [set(tuple([tuple(pp) for pp in p.points.tolist()])) for p in mpu_object.patches]
+    unique_points = []
+    overlap_points = []
+    for patch_idx,patch_i in enumerate(patch_sets):
+        patch_i_diff = patch_i.difference({})
+        patch_i_overlap = set([])
+        for patch_jdx,patch_j in enumerate(patch_sets):
+            if patch_idx == patch_jdx:
+                continue
+            else:
+                patch_i_diff = patch_i_diff.difference(patch_j)
+                patch_i_overlap.update(patch_i.intersection(patch_j))
+            print(len(unique_points))
+        unique_points.extend(list(patch_i_diff))
+        overlap_points.extend(list(patch_i_overlap))
+    unique_points = np.array(unique_points)
+    overlap_points = np.array(overlap_points)
+    fig = plt.figure()
+    ax  = fig.add_subplot(111,projection='3d')
+    ax.scatter3D(unique_points[:,0],unique_points[:,1],unique_points[:,2],color='r')
+    ax.scatter3D(overlap_points[:,0],overlap_points[:,1],overlap_points[:,2],color='b')
+    plt.show()
 #get area of contourf for L1 error estimate as a function of "tissue depth" which will basically argue
 #that for our area of interest we can ignore the problematic areas of unstable far-field growth
 
